#!/usr/local/bin/perl
#
# STAPL file editing script to generate a STAPL file capable of dynamically editing
# the bitstream data to serialize each device.
#
# Author: Ken Irving, Actel Corporation
#
# Usage:
#   keyedit4_al_new.pl <gcf> <original STAPL> <new STAPL>
#
# Note:
#
# 1. This code assumes that the BFR instances in the design are all initially
#    tied to GND.
# 2. The contraint file must contain the locations of the BFR instances, one per line;
#    placement macros are not handled by this script
# 3. The instance names of the BFR instances must contain the text DECRYPT in their names
# 4. Since the STAPL file is being modified the CRC of the file changes.  This script
#    does not recompute the new CRC, but instead sets it to 0, disabling the CRC check in
#    the player.
# 5. Since the actual editing of the bitstream data is done in the STAPL algorithm the
#    player has more work to do at runtime, which may increase programming time
# 6. BFR instances within the same parent instance are assumed to be part of the same field.
#    Each field will create a number of 32 bit variables in the resulting STAPL file
#    used to initialize the serialized fields.
#    The names of these variables is based on the hierachical path to the buffers:
#    eg 
#        DECRYPT_ONE
#
#    creates a variable
#  
#    BOOLEAN DECRYPT_ONE[32]= $000000AA
# 



sub usage {

     die "keyedit.pl  <gcf> <original STAPL> <new STAPL>";
}

$pin = 3;  # BFR
$pinMuxA = 2;  # MUX2H pin A
$pinMuxB = 4;  # MUX2H pin B

$inputColumn{2} = 3;
$inputColumn{3} = 2;
$inputColumn{4} = 2;

$invrow{2} = 31;
$invrow{3} = 26;
$invrow{4} = 31;
$noninvrow{2} = 28; 
$noninvrow{3} = 27;
$noninvrow{4} = 30;

# MUX2H pin A uses 3_26 and 4_31
$inputColumnVddl = 3;
$inputColumnGnd = 4;
$pin2vddlrow = 26;
$pin2gndrow = 31;

##################################################################
# BEGIN OF DEVICE DATA
##################################################################

##################################################################
# APA075
##################################################################

$cols_075 = 3; 
$INT_SIZE_075 = 6;

$tileSeqStr_075 = 
    "iv" .               # west io and vll
        (
            ("l" x 4) .  # 4 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 4) .  # 4 logic stiles
            "c" .        # clock spine
            ("l" x 4) .  # 4 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 4) .  # 4 logic stiles
            "v"          # extra vertical vll 
        ) x 3 .
    "i";                 # east vll and io tile


$rowSeqStr_075 = 
    "sv" .           # south io and vll row
    (
        "v" .        # vll rib
        ("l" x 8)    # 1/2 logic super duper tile
    ) x 2 .

    "g" .            # vll rib (global)

    (
        ("l" x 8) .  # 1/2 logic super duper tile
        "v"          # vll rib
    ) x 2 .

    "pc" .           # ram super duper tile
    "v" .            # vll rib
    "pc" .           # ram super duper tile
    "v" .            # vll rib
    "vn";            # vll rib and north io tile


##################################################################
# APA150
##################################################################

$cols_150 = 4; 
$INT_SIZE_150 = 7;

$tileSeqStr_150 = 
    "iv" .               # west io and vll
        (
            ("l" x 4) .  # 4 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 4) .  # 4 logic stiles
            "c" .        # clock spine
            ("l" x 4) .  # 4 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 4) .  # 4 logic stiles
            "v"          # extra vertical vll 
        ) x 4 .
    "i";                 # east vll and io tile


$rowSeqStr_150 = 
    "sv" .           # south io and vll row
    (
        "v" .        # vll rib
        ("l" x 8)    # 1/2 logic super duper tile
    ) x 4 .

    "g" .            # vll rib (global)

    (
        ("l" x 8) .  # 1/2 logic super duper tile
        "v"          # vll rib
    ) x 2 .

    "pc" .           # ram super duper tile
    "v" .            # vll rib
    "pc" .           # ram super duper tile
    "v" .            # vll rib
    "vn";            # vll rib and north io tile


##################################################################
# APA300
##################################################################

$cols_300 = 4; 
$INT_SIZE_300 = 7;

$tileSeqStr_300 = 
    "iv" .               # west io and vll
        (
            ("l" x 4) .  # 4 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 4) .  # 4 logic stiles
            "c" .        # clock spine
            ("l" x 4) .  # 4 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 4) .  # 4 logic stiles
            "v"          # extra vertical vll 
        ) x 4 .
    "i";                 # east vll and io tile


$rowSeqStr_300 = 
    "sv" .           # south io and vll row
    "v" .            # vll rib
    "pc" .           # ram super duper tile
    "v"  .           # vll rib
    "pc" .           # ram super duper tile
    (
        "v" .        # vll rib
        ("l" x 8)    # 1/2 logic super duper tile
    ) x 4 .

    "g" .            # vll rib (global)

    (
        ("l" x 8) .  # 1/2 logic super duper tile
        "v"          # vll rib
    ) x 4 .

    "pc" .           # ram super duper tile
    "v" .            # vll rib
    "pc" .           # ram super duper tile
    "v" .            # vll rib
    "vn";            # vll rib and north io tile


##################################################################
# APA450
##################################################################

$cols_450 = 6; 
$INT_SIZE_450 = 7;

$tileSeqStr_450 = 
    "iv" .               # west io and vll
        (
            ("l" x 4) .  # 4 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 4) .  # 4 logic stiles
            "c" .        # clock spine
            ("l" x 4) .  # 4 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 4) .  # 4 logic stiles
            "v"          # extra vertical vll 
        ) x 6 .
    "i";                 # east vll and io tile


$rowSeqStr_450 = 
    "sv" .           # south io and vll row
    "v" .            # vll rib
    "pc" .           # ram super duper tile
    "v"  .           # vll rib
    "pc" .           # ram super duper tile
    (
        "v" .        # vll rib
        ("l" x 8)    # 1/2 logic super duper tile
    ) x 4 .

    "g" .            # vll rib (global)

    (
        ("l" x 8) .  # 1/2 logic super duper tile
        "v"          # vll rib
    ) x 4 .

    "pc" .           # ram super duper tile
    "v" .            # vll rib
    "pc" .           # ram super duper tile
    "v" .            # vll rib
    "vn";            # vll rib and north io tile


##################################################################
# APA600
##################################################################

$cols_600 = 7; 
$INT_SIZE_600 = 8;

$tileSeqStr_600 = 
    "iv" .               # west io and vll
        (
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "c" .        # clock spine
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "v"          # extra vertical vll 
        ) x 7 .
    "i";                 # east vll and io tile


$rowSeqStr_600 = 
    "sv" .           # south io and vll row
    "v" .            # vll rib
    "pc" .           # ram super duper tile
    "v"  .           # vll rib
    "pc" .           # ram super duper tile
    (
        "v" .        # vll rib
        ("l" x 4)    # 1/2 logic super duper tile
    ) x 12 .

    "g" .            # vll rib (global)

    (
        ("l" x 4) .  # 1/2 logic super duper tile
        "v"          # vll rib
    ) x 12 .

    "pc" .           # ram super duper tile
    "v" .            # vll rib
    "pc" .           # ram super duper tile
    "v" .            # vll rib
    "vn";            # vll rib and north io tile


##################################################################
# APA750
##################################################################

$cols_750 = 8; 
$INT_SIZE_750 = 8;

$tileSeqStr_750 = 
    "iv" .               # west io and vll
        (
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "c" .        # clock spine
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "v"          # extra vertical vll 
        ) x 8 .
    "i";                 # east vll and io tile


$rowSeqStr_750 = 
    "sv" .           # south io and vll row
    "v" .            # vll rib
    "pc" .           # ram super duper tile
    "v"  .           # vll rib
    "pc" .           # ram super duper tile
    (
        "v" .        # vll rib
        ("l" x 4)    # 1/2 logic super duper tile
    ) x 16 .

    "g" .            # vll rib (global)

    (
        ("l" x 4) .  # 1/2 logic super duper tile
        "v"          # vll rib
    ) x 16 .

    "pc" .           # ram super duper tile
    "v" .            # vll rib
    "pc" .           # ram super duper tile
    "v" .            # vll rib
    "vn";            # vll rib and north io tile


##################################################################
# APA1000
##################################################################

$cols_1000 = 11; 
$INT_SIZE_1000 = 9;

$tileSeqStr_1000 = 
    "iv" .               # west io and vll
        (
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "c" .        # clock spine
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "v" .        # extra vertical vll 
            ("l" x 2) .  # 2 logic stiles
            "v"          # extra vertical vll 
        ) x 11 .
    "i";                 # east vll and io tile


$rowSeqStr_1000 = 
    "sv" .           # south io and vll row
    "v" .            # vll rib
    "pc" .           # ram super duper tile
    "v"  .           # vll rib
    "pc" .           # ram super duper tile
    (
        "v" .        # vll rib
        ("l" x 4)    # 1/2 logic super duper tile
    ) x 20 .

    "g" .            # vll rib (global)

    (
        ("l" x 4) .  # 1/2 logic super duper tile
        "v"          # vll rib
    ) x 20 .

    "pc" .           # ram super duper tile
    "v" .            # vll rib
    "pc" .           # ram super duper tile
    "v" .            # vll rib
    "vn";            # vll rib and north io tile


##################################################################
# END OF DEVICE DATA
##################################################################

# first search for the correct device then set the corresponding variables,
# for example,
#
# APA750 device
# $cols = $cols_750;
# $tileSeqStr = $tileSeqStr_750;
# $rowSeqStr = $rowSeqStr_750;
# $INT_SIZE = $INT_SIZE_750;
#

open(GCF, "< $ARGV[0]") || &usage;
open(STAPL, "< $ARGV[1]") || &usage;
open(NEW, "> $ARGV[2]") || &usage;

# default device is APA600;
$device = "APA600";
$cols = $cols_600;
$tileSeqStr = $tileSeqStr_600;
$rowSeqStr = $rowSeqStr_600;
$INT_SIZE = $INT_SIZE_600;

while (<STAPL>) {

     if (m/^\s*NOTE\s+"DEVICE"\s+"(.+)"\s*;/) {
          $device = $1;
          break;
     }
     last if /^ACTION\s+/;
}
close(STAPL);

if ($device eq "APA075") {
	$cols = $cols_075;
	$tileSeqStr = $tileSeqStr_075;
	$rowSeqStr = $rowSeqStr_075;
	$INT_SIZE = $INT_SIZE_075;
}elsif ($device eq "APA150") {
	$cols = $cols_150;
	$tileSeqStr = $tileSeqStr_150;
	$rowSeqStr = $rowSeqStr_150;
	$INT_SIZE = $INT_SIZE_150;
}elsif ($device eq "APA300") {
	$cols = $cols_300;
	$tileSeqStr = $tileSeqStr_300;
	$rowSeqStr = $rowSeqStr_300;
	$INT_SIZE = $INT_SIZE_300;
}elsif ($device eq "APA450") {
	$cols = $cols_450;
	$tileSeqStr = $tileSeqStr_450;
	$rowSeqStr = $rowSeqStr_450;
	$INT_SIZE = $INT_SIZE_450;
}elsif ($device eq "APA600") {
	$cols = $cols_600;
	$tileSeqStr = $tileSeqStr_600;
	$rowSeqStr = $rowSeqStr_600;
	$INT_SIZE = $INT_SIZE_600;
}elsif ($device eq "APA750") {
	$cols = $cols_750;
	$tileSeqStr = $tileSeqStr_750;
	$rowSeqStr = $rowSeqStr_750;
	$INT_SIZE = $INT_SIZE_750;
}elsif ($device eq "APA1000") {
	$cols = $cols_1000;
	$tileSeqStr = $tileSeqStr_1000;
	$rowSeqStr = $rowSeqStr_1000;
	$INT_SIZE = $INT_SIZE_1000;
}

@tileSeqList = split(//,$tileSeqStr);

$fc = 0;
$vllbit = $cols * 16;
$lgcbit = $cols * 16 - 1;

$x = 1;

while ($tile = shift(@tileSeqList)) { 

    if ($tile eq "i") {
        $bit = $vllbit++;
        for ($word = 0; $word < 5; $word++) { 
            $wordSelect{$fc} = $word;
            $bitSelect{$fc} = $bit;
            $fc++;
        }

    } elsif ($tile eq "l") {
        $bit = $lgcbit--;
        $tileX{$x++} = $fc;
        $tileX{$x++} = $fc+5;
      
        for( $word = 0; $word < 10; $word++) { 
            $wordSelect{$fc} = $word;
            $bitSelect{$fc} = $bit;
            $fc++;
        }

    } elsif ($tile eq "v") {
        $bit = $vllbit++;
        for ($word = 0; $word < 4; $word++) { 
             $wordSelect{$fc} = $word;
             $bitSelect{$fc} = $bit;
             $fc++;
        }
        
  } elsif ($tile eq "c") {
        $bit = $vllbit++;
        for ($word = 0; $word < 6; $word++) { 
             $wordSelect{$fc} = $word;
             $bitSelect{$fc} = $bit;
             $fc++;
        }
        
  } else {
        die "ERROR: wrong tile type\n";
  } 
}

$ws = $vllbit;


@rowSeqList = split(//,$rowSeqStr);

$lgc_tsel = 0;
$vll_tsel = @rowSeqList - 1;

$row = 0;
$y = 1;
while ($tile = shift(@rowSeqList)) { 

    if ($tile =~ /l|n|s|p/) {
    
         if ($tile eq "p") {
              $tileY{$y} = $row;
              $y = $y+2;
         } elsif ($tile eq "l") {
              $tileY{$y} = $row;
              $y++;
         }

        $tsel = $lgc_tsel++;

        for ($rsel = 0; $rsel < 32; $rsel++) { 
            $tileSelect{$row} = $tsel;
            $row++;
        }
    } elsif ($tile =~ /v|g|c/) {
        $tsel = $vll_tsel--;
        for ($rsel = 0; $rsel < 12; $rsel++) { 
            $tileSelect{$row} = $tsel;
            $row++;
        }

    } else {
        die "ERROR: wrong row type\n"
    } 
}


open(STAPL, "< $ARGV[1]") || &usage;
$firstAction = 1;

while (<STAPL>) {

     if (m/^\s*ACTION\s+/ && $firstAction) {
          print NEW "ACTION DEBUG = FIX_INT_ARRAYS, INITKEY, DUMPKEYS;\n";
          $firstAction= 0;
     }

     print NEW;
     last if /^DATA BITSTREAM/;
}

print NEW "BOOLEAN zf;\n";
# print NEW "'row = $row, y = $y\n";

$bfrKeyBits = 0;
$muxKeyBits = 0;
$bInstIsMux = 0;

while (<GCF>) {

     next if /^#/;

#    Example:  set_location (2,54) "DECRYPT_ONE/b12"; 
     if (m/^\s*set_location\s*\((\d+)\s*,\s*(\d+)\s*\)\s*(.*)\s*;/) {
#     if (m/^\s*set_initial_location\s*\((\d+)\s*,\s*(\d+)\s*\)\s*(.*)\s*;/) {
         
          $x = $1;
          $y = $2;
          $inst = $3;

          next unless $inst =~ m/DECRYPT/;
#          next unless (($inst =~ m/DECRYPT/) || ($inst =~ m/KEY_DECR/));


          $inst =~ s/^"//;
          $inst =~ s/"$//;

# This is how we change the path
# right now the syntax is DECRYPT/dat12
# DAT can be more than one character
        $inst =~ m!(.*)/[^0-9](\d+)$!;
		$key = $1;
		$bit = $2;
		$key =~ s!/!_!g;
		
        if ($inst =~ m!.*/[Mm]\d+$!) {
#        if ($inst =~ m/KEY_DECR/) {
#            print NEW "' MUX\n";
            $bInstIsMux = 1;
        }
        else {
            $bInstIsMux = 0;
        }

#        print NEW "'\n";
#        print NEW "'x = $x, y = $y inst = $inst, key=$key, bit=$bit\n";

        if ($bInstIsMux == 0) {
          $uniqueBfrKeys{$key}++;

          $bfrKeyMap{$bfrKeyBits} = "${key}[${bit}]";

          # Locate the bottom left corner of the tile

          $fc = $tileX{$x};
          $cg = $tileY{$y};
          
#         print NEW "'fc = $fc, cg = $cg\n";

          # We're using the macro BFR, which uses pin 3 (I5)
          # The grounding switch 2,28 is already on
          # For a '0' we use switch 2,27
          # For a '1' we use switch 2,26
          
          $fc = $fc + $inputColumn{$pin};
          
          # Now we know the physical address of the switches
          # so figure out the logical address.
          
          $tile = $tileSelect{$cg};
          $word = $wordSelect{$fc};
          $bit  = $bitSelect{$fc};

#          print NEW "' tile $tile, word $word, bit $bit\n";

          unshift(@tileBfrList, $tile);
          unshift(@bitBfrList, ($word*$ws+$bit));
          $bfrKeyBits++;

        }
        else {
          $bitOrig = $bit;

          $keyA = "${key}_MUXA";
          $uniqueMuxAKeys{$keyA}++;

          $muxAKeyMap{$muxKeyBits} = "${keyA}[${bitOrig}]";

          # Locate the bottom left corner of the tile

          $fcOrig = $tileX{$x};
          $cg = $tileY{$y};
          
          $tile = $tileSelect{$cg};
          unshift(@tileMuxList, $tile);

          # We're using the macro MUX, which uses pin 2 (I2) and pin 4 (I8)

          # pin 2 (I2)
          # The grounding switch 3,29 is already on
          # For a '0' set 3_26 (vddl) = 0 and 4_31 (vddl) = 1
          # For a '1' set 3_26 (vddl) = 1 and 4_31 (vddl) = 0
          
          $fc = $fcOrig + $inputColumnVddl;
          
          # Now we know the physical address of the switches
          # so figure out the logical address.
          
          $word = $wordSelect{$fc};
          $bit  = $bitSelect{$fc};

#          print NEW "' pinMuxA vddl fc = $fc, cg = $cg\n";
#          print NEW "' pinMuxA vddl tile $tile, word $word, bit $bit\n";

          unshift(@bitMuxAVddlList, ($word*$ws+$bit));

          $fc = $fcOrig + $inputColumnGnd;
          
          # Now we know the physical address of the switches
          # so figure out the logical address.
          
          $word = $wordSelect{$fc};
          $bit  = $bitSelect{$fc};

#          print NEW "' pinMuxA gnd fc = $fc, cg = $cg\n";
#          print NEW "' pinMuxA gnd tile $tile, word $word, bit $bit\n";

          unshift(@bitMuxAGndList, ($word*$ws+$bit));


          $keyB = "${key}_MUXB";
          $uniqueMuxBKeys{$keyB}++;
          $muxBKeyMap{$muxKeyBits} = "${keyB}[${bitOrig}]";

          # pin 4 (I8)
          # The grounding switch 2,29 is already on
          # For a '0' we use switch 2,30
          # For a '1' we use switch 2,31
          
          $fc = $fcOrig + $inputColumn{$pinMuxB};
          
          # Now we know the physical address of the switches
          # so figure out the logical address.
          
          $word = $wordSelect{$fc};
          $bit  = $bitSelect{$fc};

#          print NEW "' pinMuxB fc = $fc, cg = $cg\n";
#          print NEW "' pinMuxB tile $tile, word $word, bit $bit\n";

          unshift(@bitMuxBList, ($word*$ws+$bit));

          $muxKeyBits++;

        }



     }
}

print NEW "BOOLEAN rowdata[" . ($ws*10) . "];\n";
if ($bfrKeyBits > 0) {
print NEW "INTEGER BFRTILELIST[$bfrKeyBits]= " . join(',', @tileBfrList) . ";\n";
print NEW "INTEGER BFRBITLIST[$bfrKeyBits]= " . join(',', @bitBfrList) . ";\n";
}
if ($muxKeyBits > 0) {
print NEW "INTEGER MUXTILELIST[$muxKeyBits]= " . join(',', @tileMuxList) . ";\n";
print NEW "INTEGER MUXAVDDLBITLIST[$muxKeyBits]= " . join(',', @bitMuxAVddlList) . ";\n";
print NEW "INTEGER MUXAGNDBITLIST[$muxKeyBits]= " . join(',', @bitMuxAGndList) . ";\n";
print NEW "INTEGER MUXBBITLIST[$muxKeyBits]= " . join(',', @bitMuxBList) . ";\n";
}

#####################################################################
#if ($bfrKeyBits > 0) {
#
#foreach $k (keys(%uniqueBfrKeys)) {
#  $bits = $uniqueBfrKeys{$k};
#  $dwords = ($bits+31) >> 5;
#  foreach $i (0 .. $dwords-1) {
#      $var =  "${k}_" . ($i*32+31) . "_" . ($i*32);
#      push(@varList, $var);
#     print NEW "INTEGER ${var}= 61680; ' 0xF0F0\n";
#  } 
#}
#
#}
#
#if ($muxKeyBits > 0) {
#
#foreach $k (keys(%uniqueMuxAKeys)) {
#  $bits = $uniqueMuxAKeys{$k};
#  $dwords = ($bits+31) >> 5;
#  foreach $i (0 .. $dwords-1) {
#      $var =  "${k}_" . ($i*32+31) . "_" . ($i*32);
#      push(@varList, $var);
#     print NEW "INTEGER ${var}= 21845; ' 0x5555\n";
#  } 
#}
#
#foreach $k (keys(%uniqueMuxBKeys)) {
#  $bits = $uniqueMuxBKeys{$k};
#  $dwords = ($bits+31) >> 5;
#  foreach $i (0 .. $dwords-1) {
#      $var =  "${k}_" . ($i*32+31) . "_" . ($i*32);
#      push(@varList, $var);
#     print NEW "INTEGER ${var}= 43690; ' 0xAAAA\n";
#  } 
#}
#
#}
#####################################################################
if ($bfrKeyBits > 0) {

foreach $k (keys(%uniqueBfrKeys)) {
  $bits = $uniqueBfrKeys{$k};
  print NEW "BOOLEAN ${k}[${bits}]= \$" . ("0" x (($bits+3)/4)) . ";\n";
}

}

if ($muxKeyBits > 0) {

foreach $k (keys(%uniqueMuxAKeys)) {
  $bits = $uniqueMuxAKeys{$k};
  print NEW "BOOLEAN ${k}[${bits}]= \$" . ("0" x (($bits+3)/4)) . ";\n";
}
foreach $k (keys(%uniqueMuxBKeys)) {
  $bits = $uniqueMuxBKeys{$k};
  print NEW "BOOLEAN ${k}[${bits}]= \$" . ("0" x (($bits+3)/4)) . ";\n";
}

}

if ($bfrKeyBits > 0) {
print NEW "BOOLEAN BFRKEY[" . ($bfrKeyBits) . "];\n";
}
if ($muxKeyBits > 0) {
print NEW "BOOLEAN MUXAKEY[" . ($muxKeyBits) . "];\n";
print NEW "BOOLEAN MUXBKEY[" . ($muxKeyBits) . "];\n";
}

while (<STAPL>) {

    next if /^BOOLEAN zf/;

    next if /^CRC /;

   if (m/^PROCEDURE\s+(\S+)/) {
      $proc = $1;
     s/;/,BUILDROW;/ if (($proc eq "PA") || ($proc eq "V"));

        if ($proc eq "INITIALIZE") {
             s/;/,INITKEY;/;
             print NEW;
             $_ = "CALL INITKEY;\n";
        } elsif ($proc eq "FIX_INT_ARRAYS") {
             s/;/,BITSTREAM;/;
        }
  } elsif (m/^intok:/) {

     if ($bfrKeyBits > 0) {
        print NEW "FOR i=0 TO " . (($bfrKeyBits >> 1)-1) . ";\n";
        print NEW "  PUSH BFRTILELIST[i];\n";
        print NEW "  BFRTILELIST[i]=BFRTILELIST[" . ($bfrKeyBits-1) . "-i];\n";
        print NEW "  POP BFRTILELIST[" . ($bfrKeyBits-1) . "-i];\n";
        print NEW "  PUSH BFRBITLIST[i];\n";
        print NEW "  BFRBITLIST[i]=BFRBITLIST[" . ($bfrKeyBits-1) . "-i];\n";
        print NEW "  POP BFRBITLIST[" . ($bfrKeyBits-1) . "-i];\n";
        print NEW "NEXT i;\n";
     }
     if ($muxKeyBits > 0) {
        print NEW "FOR i=0 TO " . (($muxKeyBits >> 1)-1) . ";\n";
        print NEW "  PUSH MUXTILELIST[i];\n";
        print NEW "  MUXTILELIST[i]=MUXTILELIST[" . ($muxKeyBits-1) . "-i];\n";
        print NEW "  POP MUXTILELIST[" . ($muxKeyBits-1) . "-i];\n";
        print NEW "  PUSH MUXAVDDLBITLIST[i];\n";
        print NEW "  MUXAVDDLBITLIST[i]=MUXAVDDLBITLIST[" . ($muxKeyBits-1) . "-i];\n";
        print NEW "  POP MUXAVDDLBITLIST[" . ($muxKeyBits-1) . "-i];\n";
        print NEW "  PUSH MUXAGNDBITLIST[i];\n";
        print NEW "  MUXAGNDBITLIST[i]=MUXAGNDBITLIST[" . ($muxKeyBits-1) . "-i];\n";
        print NEW "  POP MUXAGNDBITLIST[" . ($muxKeyBits-1) . "-i];\n";
        print NEW "  PUSH MUXBBITLIST[i];\n";
        print NEW "  MUXBBITLIST[i]=MUXBBITLIST[" . ($muxKeyBits-1) . "-i];\n";
        print NEW "  POP MUXBBITLIST[" . ($muxKeyBits-1) . "-i];\n";
        print NEW "NEXT i;\n";
     }
  }


   if (($proc eq "PA") && m/\s*IRSCAN 8,\$e4;/) {
      print NEW;

      print NEW "  CALL BUILDROW;\n";
      print NEW "  FOR w=9 TO 0 STEP -1;\n";
        print NEW "    IRSCAN 8,\$85;\n";         # IDI
        print NEW "    DRSCAN $ws,rowdata[w*$ws+" . ($ws-1) . "..w*$ws];\n";
        print NEW "    IRSCAN 8,\$84;\n";         # ICS
        print NEW "    DRSCAN 10,BOOL(1<<w);\n";

         while (<STAPL>) {
            last if /^\s*NEXT w;/;
         } 

   } elsif (($proc eq "V") && m/^\s*FOR w=/) {
       print NEW "  CALL BUILDROW;\n";
   } elsif (($proc eq "V") && m/^\s*IRSCAN 8,\$e5/) {
       print NEW "   IRSCAN 8,\$85;\n";
       print NEW "   DRSCAN $ws,rowdata[w*$ws+" . ($ws-1) . "..w*$ws];\n";
       
       while (<STAPL>) {
         last if /^\s*IRSCAN 8,\$c8;/;
      }
   }

   print NEW;

}





print NEW "PROCEDURE BUILDROW USES GV,BITSTREAM;\n";
print NEW "zf=1;\n";
print NEW "BOOLEAN brz[" . ($ws*10) . "];\n";
print NEW "rowdata[]=brz[];\n";
print NEW "FOR w=9 TO 0 STEP -1;\n";
print NEW " IF sw[o] THEN GOTO br0;\n";
print NEW " rowdata[w*$ws+" . ($ws-1) . "..w*$ws]=sw[o+$ws..o+1];\n";
print NEW " o=o+" . ($ws+1) . ";\n";
print NEW " zf=0;\n";
print NEW " GOTO br3;\n";
print NEW " br0: o=o+1;\n";
print NEW " IF sw[o]==0 THEN GOTO br2;\n";
print NEW " zf=0;\n";
print NEW " BOOLEAN padbr[$ws];\n";
# print NEW " br1: i=INT(sw[o+8..o+1]);\n";
print NEW " br1: i=INT(sw[o+" . ($INT_SIZE) . "..o+1]);\n";
print NEW " padbr[i]=1;\n";
# print NEW " o=o+9;\n";
print NEW " o=o+" . ($INT_SIZE+1) . ";\n";
print NEW " IF sw[o] THEN GOTO br1;\n";
print NEW " rowdata[w*$ws+" . ($ws-1) . "..w*$ws]=padbr[];\n";
print NEW " br2: o=o+1;\n";
print NEW " br3:\n";
print NEW "NEXT w;\n";

if ($bfrKeyBits > 0) {
print NEW "IF ((r != $invrow{$pin}) && (r != $noninvrow{$pin})) THEN GOTO bre;\n";   # save some time
print NEW "INTEGER bri;\n";
print NEW "FOR bri=0 TO " . ($bfrKeyBits-1) . ";\n";
print NEW " IF BFRKEY[bri]==0 THEN GOTO nbri;\n";
print NEW " IF t!=BFRTILELIST[bri] THEN GOTO nbri;\n";
#print NEW " PRINT \"bri : \", bri;\n";
#print NEW " PRINT \"r : \", r;\n";
print NEW " IF r==$invrow{$pin} THEN rowdata[BFRBITLIST[bri]]=1;\n";
print NEW " IF r==$noninvrow{$pin} THEN rowdata[BFRBITLIST[bri]]=0;\n";
print NEW " zf=0;\n";
print NEW "nbri:\n";
print NEW "NEXT bri;\n";
if ($muxKeyBits > 0) {
print NEW "GOTO muxe;\n";
}
print NEW "bre:\n";
}

if ($muxKeyBits > 0) {
## $pin2gndrow == $invrow{$pinMuxB}
print NEW "IF ((r != $pin2vddlrow) && (r != $pin2gndrow) && (r != $noninvrow{$pinMuxB})) THEN GOTO muxe;\n";   # save some time
print NEW "INTEGER muxi;\n";
print NEW "FOR muxi=0 TO " . ($muxKeyBits-1) . ";\n";
print NEW " IF t!=MUXTILELIST[muxi] THEN GOTO nmuxi;\n";
#print NEW " IF ((MUXAKEY[muxi]==0) && (MUXBKEY[muxi]==0)) THEN GOTO muxab00;\n";
print NEW " IF ((MUXAKEY[muxi]==0) && (MUXBKEY[muxi]==0)) THEN GOTO nmuxi;\n";
print NEW " IF ((MUXAKEY[muxi]==0) && (MUXBKEY[muxi]==1)) THEN GOTO muxab01;\n";
print NEW " IF ((MUXAKEY[muxi]==1) && (MUXBKEY[muxi]==0)) THEN GOTO muxab10;\n";
## muxab11;
print NEW " IF r==$pin2vddlrow THEN rowdata[MUXAVDDLBITLIST[muxi]]=1;\n";
print NEW " IF r==$pin2gndrow THEN rowdata[MUXAGNDBITLIST[muxi]]=0;\n";
print NEW " IF r==$invrow{$pinMuxB} THEN rowdata[MUXBBITLIST[muxi]]=1;\n";
print NEW " IF r==$noninvrow{$pinMuxB} THEN rowdata[MUXBBITLIST[muxi]]=0;\n";
print NEW " zf=0;\n";
print NEW "GOTO nmuxi;\n";

#print NEW "muxab00:\n";
#print NEW " IF r==$pin2vddlrow THEN rowdata[MUXAVDDLBITLIST[muxi]]=0;\n";
#print NEW " IF r==$pin2gndrow THEN rowdata[MUXAGNDBITLIST[muxi]]=1;\n";
#print NEW " IF r==$invrow{$pinMuxB} THEN rowdata[MUXBBITLIST[muxi]]=0;\n";
#print NEW " IF r==$noninvrow{$pinMuxB} THEN rowdata[MUXBBITLIST[muxi]]=1;\n";
#print NEW " zf=0;\n";
#print NEW "GOTO nmuxi;\n";

print NEW "muxab01:\n";
#print NEW " IF r==$pin2vddlrow THEN rowdata[MUXAVDDLBITLIST[muxi]]=0;\n";
#print NEW " IF r==$pin2gndrow THEN rowdata[MUXAGNDBITLIST[muxi]]=1;\n";
print NEW " IF r==$invrow{$pinMuxB} THEN rowdata[MUXBBITLIST[muxi]]=1;\n";
print NEW " IF r==$noninvrow{$pinMuxB} THEN rowdata[MUXBBITLIST[muxi]]=0;\n";
print NEW " zf=0;\n";
print NEW "GOTO nmuxi;\n";

print NEW "muxab10:\n";
print NEW " IF r==$pin2vddlrow THEN rowdata[MUXAVDDLBITLIST[muxi]]=1;\n";
print NEW " IF r==$pin2gndrow THEN rowdata[MUXAGNDBITLIST[muxi]]=0;\n";
#print NEW " IF r==$invrow{$pinMuxB} THEN rowdata[MUXBBITLIST[muxi]]=0;\n";
#print NEW " IF r==$noninvrow{$pinMuxB} THEN rowdata[MUXBBITLIST[muxi]]=1;\n";
print NEW " zf=0;\n";

print NEW "nmuxi:\n";
print NEW "NEXT muxi;\n";
print NEW "muxe:\n";
}

print NEW "ENDPROC;\n\n";





print NEW "PROCEDURE INITKEY USES BITSTREAM;\n";


if ($bfrKeyBits > 0) {

$bfrKeyNum = keys(%bfrKeyMap);
# print NEW "$bfrKeyNum\n";
foreach $i (0 .. $bfrKeyNum-1) {
  print NEW "BFRKEY[$i]=$bfrKeyMap{$i};\n";
}

}


if ($muxKeyBits > 0) {

$muxAKeyNum = keys(%muxAKeyMap);
# print NEW "$muxAKeyNum\n";
foreach $i (0 .. $muxAKeyNum-1) {
  print NEW "MUXAKEY[$i]=$muxAKeyMap{$i};\n";
}

$muxBKeyNum = keys(%muxBKeyMap);
# print NEW "$muxBKeyNum\n";
foreach $i (0 .. $muxBKeyNum-1) {
  print NEW "MUXBKEY[$i]=$muxBKeyMap{$i};\n";
}

}

print NEW "ENDPROC;\n\n";

print NEW "PROCEDURE DUMPKEYS USES BITSTREAM;\n";

#####################################################################
#foreach $var (@varList) {
#  print NEW "PRINT \"$var = \", $var;\n";
#}
#####################################################################

if ($bfrKeyBits > 0) {

foreach $k (keys(%uniqueBfrKeys)) {
  $bits = $uniqueBfrKeys{$k};
  print NEW "EXPORT \"${k}\", ${k}[];\n"; 
}

}

if ($muxKeyBits > 0) {

foreach $k (keys(%uniqueMuxAKeys)) {
  $bits = $uniqueMuxAKeys{$k};
  print NEW "EXPORT \"${k}\", ${k}[];\n"; 
}
foreach $k (keys(%uniqueMuxBKeys)) {
  $bits = $uniqueMuxBKeys{$k};
  print NEW "EXPORT \"${k}\", ${k}[];\n"; 
}

}

print NEW "EXIT 0;\n";
print NEW "ENDPROC;\n";


print NEW "CRC 0;\n";   # disable CRC check


close(NEW);
close(STAPL);
close(GCF);



